*********************************************************************;
******************** META-ANALYSIS WITH SAS *************************;
************** (comments and notes by L. V. Madden) *****************;
************************ Ohio State University **********************;
*********************************************************************;

/*
Demonstration of various types of meta-analyses.
The macros come directly from Lipsey and Wilson book (and website), 
with subsequent revisions. (Macros included here by permission of D. Wilson).
The use of MIXED comes from various sources, especially van Houwelingen et al. (2002).

Scroll below the macro listings to see data and example analyses....
Comments are given throughout. Lots of tricks are given, and some explanations.

Code prepared for instructional purposes. I make no warranties regarding the 
accuracy of all the statements or the output. 
*/

%MACRO metareg(es,w,x,dsn=_last_,model=fe) ;
*--Revised Lipsey-Wilson macro, 7/2005 (L. Madden). This version CORRECTLY does the ML estimation;
      proc iml;

        use &dsn ;
        read all var {&es} into es  where(&es^=. & &w^=. & &x^=.) ;
        read all var {&w}  into w   where(&es^=. & &w^=. & &x^=.) ;
        read all var {&x}  into x   where(&es^=. & &w^=. & &x^=.) ;

        k = nrow(es)  ;
        df = k - 1   ;
        j = j(k,1,1) ;
        x = j(k,1,1) || x ;
        p = j(1,ncol(x),1) ;
		v = 1/w;

        %if %upcase(&model) = MM | %upcase(&model) = ML %then
           %do;
                  xw = x#(w*p) ; xwx = T(xw)*x ;
                  B = inv(xwx)*T(xw)*es ;
                  qe = sum(es#w#es) - T(B)*xwx*B ;
                  ww = w#w ; xww = x#(ww*p) ; xwwx = T(xww)*x ;
                  c = (qe-(k-ncol(x)))*inv(sum(w) - trace((xwwx)*inv(xwx))) ;
                  if c<0 then c = 0 ;
                  w = 1/(1/w + c) ;
            %end ;

        %if %upcase(&model) = ML %then
           %do i=1 %to 100;
                          xw = x#(w*p) ; xwx = T(xw)*x ;
                          B = inv(xwx)*T(xw)*es ;
                          r = es - x*B ;
                          c = sum((w#w)#(r#r-(v)))/
                                (sum(w#w)) ;
                          if c<0 then c = 0 ;
                          w = 1/(v+c) ;

			 %end ;

		%if %upcase(&model) = ML %then
			%do ;  se_c = sqrt(2/sum(w#w)) ; %end ;

        xw = x#(w*p) ; xwx = T(xw)*x ;
        B = inv(xwx)*T(xw)*es ;

        meanes = sum(es#w)/sum(w) ;
        q =  sum((meanes - es)#(meanes - es)#w) ;
        qe = sum(es#w#es) - T(B)*xwx*B ;
        qr = q - qe ;
        dfe = nrow(es)-ncol(x) ;
        dfr = ncol(x)-1 ;
        dft = nrow(es)-1 ;
        pt = 1 - probchi(q,dft)  ;
        pe = 1 - probchi(qe,dfe) ;
        pr = 1 - probchi(qr,dfr) ;
        se = sqrt(vecdiag(inv(xwx))) ;
        zvalues = B#vecdiag(inv(diag(se))) ;
        pvalues = (1 - probnorm(abs(zvalues)))*2 ;
        lowerB = B - se*1.96 ;
        upperB = B + se*1.96 ;

        d = x - T(T((x#(w*p))[+,])*T(j)/sum(w)) ;
        sx = sqrt(vecdiag(T(d#(w*p))*d/sum(w))) ;
          sy = sqrt(q/sum(w)) ;
        beta = (B#sx) *inv(sy) ;
        r2 = qr/(qr+qe) ;
              %if %upcase(&model) = MM %then
                %do ; print "--------- Random Effects Model ------" ;
                      print "-- estimated via method of moments --" ;
                print c [label= ' '] [rowname = 'Random eff. var. component = ']
                                [format=12.8];
                %end ;

              %if %upcase(&model) = ML %then
                %do ; print "--------- Random Effects Model ------" ;
                      print "-- estimated via iterative full information maximum likelihood --" ;
                mlc = c // se_c ;
				print mlc [label= ' '] [rowname = {'Random eff. var. component  = '
				                                  'St. Error of var. component = '}]
                                [format=12.8];
                %end ;


              print "------- Descriptives -------" ;
              print k      [label="No. of Obs. "]
                         meanes [label="Mean Obs."] [format=10.4]
                        r2     [label="R-squared"] [format=10.4] ;

              print  "------- Homogeneity Analysis  -------";
                qtab1 = qr // qe // q ;
                qtab2 = dfr // dfe // dft ;
                qtab3 = pr // pe // pt ;
                print qtab1 [rowname={"Model","Residual","Total"}]
                            [colname="Q"][label="Source"] [format=12.5]
                            qtab2 [label="df"]
                            qtab3 [label="p"] [format=12.5];

              print "------- Regression Coefficients -------" ;
              names = {&x} ;
              names = T(names) ;
              group = {"Constant"} // names ;
              print b [rowname=group] [colname="B"] [label="Variable"]
                          [format=10.5]
                        se [label="SE"] [format=10.5]
                        lowerB [label='-95%CI'] [format=10.5]
                        upperB [label='+95%CI'] [format=10.5]
                        zvalues  [label="z"] [format=10.5]
                        pvalues  [label="p"] [format=10.5]
                        beta     [label="beta"] [format=10.5] ;

%MEND metareg ;

run; quit;

%MACRO meanes(es,w,dsn=_last_,print=raw) ;	
		*---Lipsey-Wilson macro for mean effect size and other statistics,
			based on moment-type calculations;
        proc iml;
                use &dsn ;
                read all var{&es} into es   where(&es^=. & &w^=.) ;
                read all var{&w}  into w    where(&es^=. & &w^=.) ;

                k = nrow(es)  ;
                df = k - 1   ;

                mes = sum(es#w)/sum(w);
                sem = sqrt(1/sum(w));
                les = mes - 1.95996*sem;
                ues = mes + 1.95996*sem;
                z   = mes/sem ;
                pz  = (1 - (.5+erf(abs(z)/sqrt(2))/2))*2 ;
                maxes = max(es) ;
                mines = min(es) ;

                q = sum((es#es)#w) - sum(es#w)#sum(es#w)/sum(w);
                wsd   = sqrt(q*w[+,]**-1) ;
                pq   = 1-probchi(q,df) ;

                c    = (q - df)/(w[+,]-sum(w#w)/w[+,]) ;
                if c<0 then ; do ; c = 0 ; end ;
                wre    = 1/(1/w + c) ;

                mesre = sum(es#wre)/sum(wre) ;
                semre = sqrt(1/sum(wre)) ;
                lesre = mesre -    1.95996*semre ;
                uesre = mesre +    1.95996*semre ;
                zre   = mesre/semre ;
                pzre  = (1 - (.5+erf(abs(zre)/sqrt(2))/2))*2 ;

                %if %upcase(&print) = EXP %then
                        %do;
                                mes = exp(mes) ;
                                les = exp(les) ;
                                ues = exp(ues) ;
                                sem = . ;
                                mes_re = exp(mes_re) ;
                                les_re = exp(les_re) ;
                                ues_re = exp(ues_re) ;
                                semre = . ;
                %end;

                print '-------------- Distribution Description --------------';
                mattrib         k     label={"No. of obs."}     maxes label={"Max Obs."}
                        mines label={"Min Obs."} wsd   label={"Weighted SD"};
                print k mines maxes wsd [format=12.5];
                print " ";
                print '-------------- Homogeneity Analysis --------------';
                mattrib df label={"df"} pq label={"p"} ;
                print q [format=12.5] df pq [format=12.5];
                print " ";
                print '-------------- Fixed & Random Effects Model --------------';
                fixed = mes || sem || les || ues || z || pz ;
                random = mesre || semre || lesre || uesre || zre || pzre ;

                model = fixed // random ;
                mattrib model rowname=({'Fixed', 'Random'})
                        colname=({'Mean' 'SE' '-95%CI' '+95%CI' 'z' 'p'})
                        label={"Model"};
                print model [format=10.5];
                mattrib c label={" "};
                print c [rowname="Random effects var. component = "];

%MEND meanes;
run; quit;


%MACRO metaf(es,w,x,dsn=_last_,model=fe) ;
*--revised Lipsey-Wilson macro , 7/05 (L Madden). Does moment or ML estimation;
      proc iml;

        use &dsn ;
        read all var {&es} into es  where(&es^=. & &w^=. & &x^=.) ;
        read all var {&w}  into w   where(&es^=. & &w^=. & &x^=.) ;
        read all var {&x}  into grp where(&es^=. & &w^=. & &x^=.) ;

        k = nrow(es)  ;
        df = k - 1   ;
        x = design(grp) ;
        p = j(1,ncol(x),1) ;
        j = j(k,1,1) ;
        v = 1/w ;

        %if %upcase(&model) = MM | %upcase(&model) = ML %then
           %do;
                  xw = x#(w*p) ; xwx = T(xw)*x ;
                  B = inv(xwx)*T(xw)*es ;
                  qe = sum(es#w#es) - T(B)*xwx*B ;
                  ww = w#w ; xww = x#(ww*p) ; xwwx = T(xww)*x ;
                  c = (qe-(k-ncol(x)))*inv(sum(w) - trace((xwwx)*inv(xwx))) ;
				  
                  if c<0 then c = 0 ;
                  w = 1/(v + c) ;
            %end ;

        %if %upcase(&model) = ML %then
           %do i=1 %to 100;
                          xw = x#(w*p) ; xwx = T(xw)*x ;
                          B = inv(xwx)*T(xw)*es ;
                          r = es - x*B ;
                          c = sum((w#w)#(r#r-(v))) /
                                (sum(w#w)) ;
                          if c<0 then c = 0 ;
			  w = 1/(v+c) ;
           %end ;

	%if %upcase(&model) = ML %then
			%do ;  se_c = sqrt(2/sum(w#w)) ; %end ;

	  group = char(inv(T(x)*x)*T(x)*grp,8,0) ;
	  means = T(T(T(x#(w*p))*es)*inv(T(x#(w*p))*x)) ;
	  grpns = vecdiag(T(x)*x) ;
	  q =  T(x#(w*p))*(es##2)-T(T((T(x#(w*p))*es)##2)*inv(T(x#(w*p))*x)) ;
	  qt = sum(es##2#w) - sum(es#w)**2/sum(w) ;
	  qw = sum(q) ;
	  qb = qt - qw ;
	  dfb = ncol(x) - 1 ;
	  dfw = nrow(es) - ncol(x) ;
	  dft = nrow(es) - 1 ;
	  pb   = 1 - probchi(qb,dfb) ;
	  pw   = 1 - probchi(qw,dfw) ;
	  ptot = 1 - probchi(qt,dft) ;
	  se = sqrt(vecdiag(inv(T(x#(w*p))*x)));
	  tvalues = means#vecdiag(inv(diag(se))) ;
	  pt = 1-probf(tvalues#tvalues,1,dfw) ;
	  lmeans = means - (se#1.96) ;
	  umeans = means + (se#1.96) ;
	  gmean = sum(es#w)/sum(w) ;
	  segmean = sqrt(1/sum(w)) ;
	  df = grpns - 1 ;    
	  pq = 1 - probchi(q,df) ;

      %if %upcase(&model) = MM %then
        %do ; print "--------- Random Effects Model ------" ;
              print "-- estimated via method of moments --" ;
        print c [label= ' '] [rowname = 'Random eff. var. component = ']
                        [format=12.8];
        %end ;

      %if %upcase(&model) = ML %then
        %do ; print "--------- Random Effects Model ------" ;
              print "-- estimated via iterative full information maximum likelihood --" ;
        mlc = c // se_c ;
		print mlc [label= ' '] [rowname = {'Random eff. var. component  = '
                  'St. Error of var. component = '}]
                  [format=12.8];
        %end ;

	  print  "------- Analog ANOVA table (Homogeneity Q)  -------";
	  qtab1 = qb // qw // qt ;
	  qtab2 = dfb // dfw // dft ;
	  qtab3 = pb // pw // ptot ;
	  print qtab1 [rowname={"Between","Within","Total"}] 
	  			  [colname="Q"]		[label="Source"] [format=12.5]
	  		qtab2 [label="df"] 
	  		qtab3 [label="p"] [format=12.5];
	
	  print "------- Q by Group -------" ;
	  print q [rowname=group] 
	  			  [colname="Q"]	[label="Group"] [format=12.5]
		 	df [label="df"]
			pq [label="p"] [format=12.5];
	  
	  print "------- Effect Size Results by Group -------" ;
	  means = means // gmean ;
	  se = se // segmean ;
	  lmeans = lmeans // gmean-(segmean*1.96) ;
	  umeans = umeans // gmean+(segmean*1.96) ;
	  tvalues = tvalues // gmean/segmean ;
	  pt = pt // 1-probf((gmean/segmean)#(gmean/segmean),1,dfw) ;
	  grpns = grpns // k ;
	  group = group // "   Total" ;
	  print means [rowname=group] [colname="Means"] [label="Group"]
	  		[format=10.5]
			se [label="SE"] [format=10.5]
			lmeans [label='-95%CI'] [format=10.5]
			umeans [label='+95%CI'] [format=10.5]
			tvalues  [label="t"] [format=10.5]
			pt  [label="p"] [format=10.5]
			grpns [label="n"];

%MEND metaf;
run;quit;


/*
		**** UNIVARIATE EXAMPLE (single effect size per study) ****

Demo with simulated data (50 studies).
efft:		effect size per study;
var_efft:	variance of the effect size per study (square of SE of effect size);
moder1:		a categorical moderator variable (e.g., cultivars 1, 2, 3)'
moder2:		a continuous moderator variable (e.g., avg. T);
*/

data demo;								*<--simulated data;
input moder1 moder2 efft var_efft studyno;
wgt = 1/var_efft;						*<--for macros and PROCs, need weight per study (1/var);
zzz = moder2-moder2 + 1;				*<--a null variable for a trick below;
datalines;
1	21	1.13286	20.1732	1
0	19	0.91970	1.4405	2
1	22	0.72507	5.8174	3
1	24	-0.47340	24.8908	4
1	24	0.56236	2.4292	5
2	21	3.43668	6.4030	6
1	15	1.60458	2.1051	7
0	15	1.42985	1.6785	8
0	18	1.58239	1.5939	9
0	20	1.92021	0.0963	10
1	21	1.95475	3.0475	11
1	18	1.18504	0.4636	12
1	22	1.16712	1.1695	13
0	18	0.70154	1.3622	14
0	25	0.67354	0.8050	15
0	19	0.99711	2.0288	16
0	23	2.27181	5.1034	17
2	22	2.12609	23.5246	18
2	21	2.76217	1.9400	19
2	18	1.36925	0.7100	20
2	17	0.46791	3.3591	21
1	20	0.86908	4.0169	22
2	21	1.45729	18.6656	23
2	18	2.64130	1.8637	24
0	19	1.76658	0.3448	25
0	15	0.69369	4.2905	26
0	18	1.61441	0.0928	27
2	23	2.63756	22.3204	28
0	17	1.50678	0.2252	29
0	21	-0.11565	0.1711	30
0	15	0.05201	0.0692	31
2	22	-1.39402	3.9929	32
2	23	1.35610	1.0142	33
1	21	1.10818	5.0666	34
0	15	0.22629	0.9755	35
2	16	-0.03458	15.4405	36
0	20	1.87402	1.0009	37
1	19	2.54478	20.9345	38
1	19	1.28713	0.4589	39
1	16	0.22484	3.7634	40
0	21	3.02889	1.6573	41
2	20	0.55123	2.6158	42
0	22	-0.74106	0.2730	43
2	19	1.71147	1.2556	44
2	18	0.43599	9.7652	45
1	20	1.20071	0.9767	46
2	25	2.06417	6.0231	47
0	18	1.35886	0.6265	48
1	26	0.75691	12.5355	49
0	19	0.96668	0.3826	50
;

title 'Demo of univariate meta-analysis with macros or SAS procedures'; 
footnote;

*ods html style=banker path='C:\temp2' (URL=NONE)  body='sastest.html';

*---If you use these macros, please cite the book:
	Lipsey, M.W., and Wilson, D. B. 2001. Practical Meta-Analysis. SAGE Publ., 
	Thousand Oaks, USA.;

*---macros use positional arguments (first two or three), and explict (=) arguments;
*---first give effect size/study, weight (1 over sampling variance per study);
*---then give file name.;
*---this macro does strictly MOMENT estimates of the mean and among-study variance;

*---macros are a bit crude, but work for many circumstances;

title2 'Moment-based univariate analysis (showing fixed and random-effect estimates';
title3 '(Lipsey & Wilson book, plus macro revisions)';
%meanes(efft,wgt,dsn=demo,print=raw) ;
run;

*---there is a trick to get ML estimates with no moderator variables with a macro;
*---first calculate a variable with ALL 0s (or 1s, etc.) in a data step;
*---then use the macro for a class moderator variable (with the null variable
	inserted for the moderator;
*---the test for the moderator here is meaningless, but the estimated variance is OK;
*---(the system log indicates an error, but all is OK);

title2 'Trick to get ML estimates (random effect) using class-moderator macro';
title3 '(Lipsey & Wilson book, plus macro revisions)';
%metaf(efft,wgt,zzz,dsn=demo,model=ml) ;
run;

*---in above (and below), the revised ML algorthim in the macro is rather crude (but often works);
*---I believe the initial estimate of the between-study variance comes from the;
*---moment estimate of this variance. If that estimate is 0 (easily possible);
*---then the search algorithm could get stuck, and not find the ML estimate;
*---The estimate of the SE(var est.) is based on the _Expected_ Fisher Information, not
	the actual (estimated) Information;

*---now do a ML-based analysis of the class moderator MODER1;

title2 'ML-based analysis of class moderator variable';
title3 '(Lipsey & Wilson book, plus macro revisions)';
%metaf(efft,wgt,moder1,dsn=demo,model=ml) ;
run;

*---now do a ML-based analysis of the continuous moderal MODER2;
*---in output, "B" is the estimated coefficient, and "beta" is the standardized coef.;

title2 'ML-based analysis of the continous moderator MODER2';
title3 '(Lipsey & Wilson book, plus macro revisions)';
%metareg(efft,wgt,moder2,dsn=demo,model=ML) ;
run;

/*
For the most part, ML or REML fitting is done from this point on, utilizing MIXED 
(or GLIMMIX). However, it must be pointed out that moment-based methods remain
very popular, especially for univariate meta-analyses. A hallmark of the moment-based
approach is the _Q_ statistic of Cochran. This is really a weighted sum of squares about
the estimated expected value, where the latter is based on a FIXED-effects
analysis (see results above for the first use of %meanes(.) above). Q is not obtained in 
a likelihood-based analysis. One can obtain Q from a fixed-effects analysis using the
GLM procedure (and weights). Almost all the output from GLM is incorrect, however, because 
one _cannot_ fix the residual variances for each study at the required values 
(see below for discussion of this idea). The _only_ relevant and correct piece of
output from GLM is the Error Sum of Squares, which is the same as Q. The following
code obtains Q from a run of GLM (based on weighted least squares), but only displays
a subset of the results. Ignore everything but this one statistic. Then, the data step
calculates the P value for the chi-squared test of homegeneity, and also calculates
two indices of Higgins and Thompson.
*/
proc glm data=demo;
title2 "Cochran's Q, df, test of homogeneity, and Higgins-Thompson H2 & I2";
ods listing select overallanova;		*<--look only at a table that contains SSE;
ods output overallanova=q_out;			*<--store Q (=SSE) for manipulation below;
class studyno;
weight wgt;
model efft = / solution;
run;
data q_out2; set q_out(where=(Source='Error'));	*<--manipulate the Q value;
	Q = ss;
	H2 = Q/df;					*<--H-squared statistic of Higgins and Thompson;
	I2 = 100*(H2-1)/H2;			*<--I-squared statistic of Higgins and Thompson;
	Pq = 1 - probchi(Q,df);		*<--P value for test of homogeneity (Cochran Q test);
	keep Q  df Pq H2 I2 I22;
proc print data=q_out2;
	var Q df Pq H2 I2 ;
run;

/*
The P value (Pq=0.134) that is displayed is not significant here. But note, the
chi-squared test is very conservative, meaning that it is hard to declare 
significance even when the variance is truly greater than 0. 

One can get the MOMENT-based estimate of the among-study variance using the
following code, which uses the value of Q calculated in the above file.
*/

data demo2; set demo; 				*<--temporary file;
wgt2=wgt**2;						*<--get weights squared (same as sampling SD to 4th power);
proc means data=demo2 sum noprint;	*<--get and store the sum of weights and weights-squared;
var wgt wgt2;
output out=demo2out sum(wgt)=wgts sum(wgt2)=wgt2s;
run;
data q_out3; merge q_out2 demo2out;	*<--combine/merge files (Q, etc., with sum of weights);
data q_out3; set q_out3;			*<--get c coefficient (see Appendix in Madden & Paul);
c = wgts - (wgt2s/wgts);
variance_MM = (Q - df)/c;			*<--get moment-based among-study variance estimate;

title2 'Moment-based estimate of the among-study variance (cf. macro results)';
proc print data=q_out3;
var Q c variance_MM;
run;

*---In the example, the moment-based estimate of the variance is about 0.15, this 
	is only about half the ML-based estimate. I prefer ML (or REML) analyses.;

*---Note: the macros above base results on ch-square tests for effects, and standard normal
	for confidence intervals, etc. In other words, same as F or student t with ddf=infinity;
*---One can force large df in SAS MIXED, using ddf= options;
*---With large ddf, one gets the same results for F (or t) test and with chi-square (z) tests;

*---now use PROC MIXED for meta-analysis, likelihood (ML) based; 
*---Procedures does Wald-type tests (and macros do sum of squares);

*---NOTE: ML (or REML) will give different results (for means and SEs) than obtained
	with the method-of-moments. Especially note: the estimate of the among-study 
	variance with the example is very different between the ML method (about 0.33)
	and the moment estimate (about 0.15).;


title2 'ML analysis in SAS, RANDOM effects, no moderators';
proc mixed data=demo method=ml covtest;
class studyno;
weight wgt;								*<--weight is inverse of sampling variance per study;
model efft = / chisq s df=10000;		*<--model gives just the global mean (intercept);
random studyno;							*<--define a random study effect (get variance);
repeated;								*<--this and next line force fixed sampling variances;
parms (1) (1) / eqcons=2;				*<--forces the sampling-variance to be 1*(1/weight);
estimate 'meanES' int 1 / cl df=10000;	*<--another way of displaying est. expected effect size;
run;

/*
Note: In above, a 'trick' is used to fix the so-called (within-study) 'sampling' variances
at pre-determined values.
This is done by forcing the residual to be '1' for all studies, *and* to simultaneously fix
the WEIGHT (which is really a within-study weight) as 1/(sampling variance). Here, the
weight is called wgt. In the PARMS statement, there are two terms: the first one is the
initial guess of the among-study variance, and the second is the initial (and final)
choice of the within-study sampling variance (1), which is held constant with the 
eqcons=2 option (not allowing the second parameter to vary). 

In the output, the among-study variance estimate is given in the table of Covariance
Parameter Estiamtes (studyno). The same table gives an 'estimate' of 1 for the 
Residual variance. But, because there are weights (that vary with study), this really
means that there is a different fixed sampling variance for each study [1*(1/weight)].

Below, the harder way to do this is shown, which is often given in meta-papers.
*/


*---The next part is another way to do the same analysis (more tedious);
data var1; set demo;
est = var_efft;		*<--define the sampling variances as ESTimates of the residual variances;
keep est studyno;
data var2;
est = 1;			*<--define the starting guess for the among-study variance;
data variance; set var2 var1;	*<--stack the variance files (among-study, then all
									the within-study sampling variances;
run;
*---In above, a special (Parms Data) file was made, which starts with an arbitrary
	first guess at the among-study variance, followed by the 50 (here) sampling
	variances (one from each study). It is a good idea to have the original data
	file sorted by study (studyno here) before doing any of this;

title2 'a look at the guess of among-study var., and all the sampling variances';
title3 '(the sampling variances will not change in the model fitting)';
proc print data=variance;
run;
title2 'ML analysis in SAS, RANDOM effects (use of PDATA= for sampling variances)';
proc mixed data=demo method=ml covtest;
class studyno;
model efft = / chisq s df=10000;			*<--note: *no* weight;
random studyno;
repeated / group=studyno type=simple;
parms / pdata=variance eqcons=2 to 51;		*<--PDATA= gives the variances;
											*^--All but the first are held constant;
estimate 'meanES' int 1 / cl df=10000;
run;
*---In the output for the PDATA approach, the table of Covariance Parameter Estimates
	first gives the among-study variance estimate (.3272), identical to the result with
	the earlier MIXED run (using the WEIGHTS and =1 for residual), followed by the
	fixed sampling variances. These are the same as those listed in the Variance
	data file.; 
*---Note in the above, the model fit results are identical (meanES, standard error, 
	likelihood, etc.) to the results obtained with the easier-to-use WEIGHTS approach.;
*---An even more tedious approach is to list all the sampling variances directly
	in a PARMS statement. This is not shown here.;

*---Now we go back to using the combination of WEIGHT and fixing the (single) residual 
	variance at 1 (which really gives fixed separate sampling variances for different
	studies...;
*---Use the following to obtain a fixed (common) effect analysis with MIXED.;

title2 'ML analysis in SAS, FIXED (common) effects, no moderators';
proc mixed data=demo method=ml covtest;
class studyno;
weight wgt;
model efft = / chisq s df=10000;
repeated;					*<--note: no random effects;
parms (1) / eqcons=1;		*<--forces sampling-variance to be 1*(1/weight) ;
estimate 'meanES' int 1 / cl df=10000;
run;

*---Note in above: one cannot get both fixed- and random-effects results from a single run
	of PROC MIXED. One must do two separate runs. Here the fixed-approach is taken. There
	is only the single (residual) variance term (fixed at 1). Coupled with differential
	weighting for the studies, and the use of the PARMS statement, 
	the mean effect size is determined without estimating any variance terms.;
*---The R2 statistic of Higgins & Thompson is obtained by dividing the SE of the random-
	effect effect size by the SE of the fixed-effect effect size, and squaring this.
	For the example: R2 = (0.1731/0.1129)^2 = 2.3;
*---The R2 statistic is _not_ a coefficient of determination;
	

*---Effect of moder1 moderator variable...;
*---See if MODER1 affects the effect size;

title2 'ML analysis, RANDOM effects of study, CLASS moderator';
proc mixed data=demo method=ml covtest;
class studyno moder1;
weight wgt;
model efft = moder1 / chisq s df=10000;
random studyno;
repeated;
parms (1) (1) / eqcons=2;
lsmeans moder1 / cl df=10000;				*<--get means for each group;
run;


*---One more demo of use of PDATA= to fix sampling variances. This is all easier
	with WEIGHTS (see above).;
title2 'ML analysis, RANDOM effects, CLASS moderator, (use of PDATA=)';
proc mixed data=demo method=ml covtest;
class studyno moder1;
model efft = moder1 / chisq s df=10000;
random studyno;
repeated / group=studyno type=simple ;
parms / pdata=variance eqcons=2 to 51;		*<--file with initial variance estimates;
											*^--first one is among-study, rest are sampling var.;
											*^--fix all but the first one;
lsmeans moder1 / cl df=10000;				*<--Get means for each group;
run;


title2 'ML analysis, RANDOM effects, CONTINUOUS moderator';
proc mixed data=demo method=ml covtest;
class studyno moder1;
weight wgt;
model efft = moder2 / chisq s df=10000;
random studyno;
repeated;
parms (1) (1) / eqcons=2 ;
run;

****************************************************************************;
************* MULTIVARIATE ANALYSES (MIXED TREATMENT COMPARIONS) ***********:
****************************************************************************;

/*
Now suppose there are three treatments in each study, called a, b, and c.
Also suppose that one has the estimated 'sampling' (i.e., within-study) variance (s2) for the
means of each study (sampling variance same for each treatment mean here).
s2 is not the variance of the mean difference (which is often available from
studies, through the LSD or MSE, etc.). I could do the following with the latter
situation, by flipping some things around. In example, I assume a single s2 for
each study.

For clarification, suppose V is the Mean Square Error (MSE) for a study with 
a one-way layout. Then the (estimated) variance of a mean is s2 = V/n, where n
is the number of replicates. Sqrt(s2) is the SE for the mean. 

To do a univariate analysis, you would need as the effect size (es) the difference
in means for each study (e.g., a-b), and the sampling variance for this 
effect size (here, s2 + s2, assuming independent means). Of couse, one would have
three separate analyses (a-b, b-c, and a-c). The differences of the es's would
be correlated within studies (same mean is used in more than one difference).
Thus, the univariate approach, although fine, is not the most efficient, and
does not take into account the correlations of effect sizes among studies.
I demonstrate the univariate approach for es = a-b below.

Note: Below I reverted to the use of REML (not ML). Thus, I also did not 
override the df values. This can certainly all be done, if desired.
*/

title 'Demo of univariate vs. multivariate approaches';
data three;
input study a b c s2 moder1;
es = a - b;
esbc = b - c;
esac = a - c;
samplingv = s2 + s2;
samplewgt = 1/samplingv;
datalines;
1 20 15  30 2.5 1
2 40 10 40 6 1
3 10 20 22 4.5 2
4 40 25 47 .5 2
5 15  5  28 2 1
6 30 35 27 3 2
7 25 13 20 1 1
8 16  8 21 .5 2
9 30 31  32 .5 2
10 12 10 9 1 1
11 19 18 24 .5 2
12 25 20 15 1.5 1
13 10 1  27 5 2
14 0 1 3   .5 2
15 6 3 16    2 1
;
run;

data threestack; set three;		*<--Create stacked file for multivariate-type analysis;
y = a;
treat='a';
samplingv=s2;			*<--In this example, sampl. var. is same for each mean;
moder=moder1;			*<--Sampling var. could vary with means within studies also;
output;
y = b;
treat = 'b';
samplingv=s2;
moder=moder1;
output;
y = c;
treat = 'c';
samplingv=s2;
moder=moder1;
output;
run;
*^--In above, one gets three records for each study (one for each treatment);

data threestack; set threestack;
samplewgt = 1/samplingv;			*<--weight for mult.-type analysis;

****************************************************************************;

title2 'Stacked data for doing mult. var. meta-analysis';
proc print data=threestack;run;
/*One can start the multivariate analysis by directly entering the data for 
	study, samplewgt, and es 
directly into a file (as constructed for threestack here).
The many manipulations above were to allow easy univariate analyses, followed by
the multivariate analysis (below).*/

****************************************************************************;

run;
title2 'univariate analysis of a-b (part of 3-treat. situation)';
proc mixed data=three method=reml;
class study;
weight samplewgt;
model es = / chisq s;
random int / sub=study;
repeated;
parms (1) (1) / eqcons=2;
run;

/*
Note: in above, an alternative version of the RANDOM statment is used:
	random int / sub=study;
This means an "intercept" (all 1s), with study as the "subject". 
This is the same as:
	random study;
Both give identical results. The explicit subject designation can make it
easier to see the generalizations for more complex scenarios. 
*/

title2 'univariate analysis of b-c (part of 3-treat. situation)';
proc mixed data=three method=reml;
class study;
weight samplewgt;
model esbc = / chisq s;
random int / sub=study;
repeated;
parms (1) (1) / eqcons=2;
run;
title2 'univariate analysis of a-c (part of 3-treat. situation)';
proc mixed data=three method=reml;
class study;
weight samplewgt;
model esac = / chisq s;
random int / sub=study;
repeated;
parms (1) (1) / eqcons=2;
run;


/*
Below is one type of multivariate analysis (multiple-treatment analysis)
that can be done. It is not the only possibility. 
It is one type of generalization of the univariate approach.
It is not the same as doing three separate univariate analyses, because the 
among-study correlations of y influence the results here (where these 
correlations cannot be utilized univariate analyses. Although the data are 
complete here, this need not be the case. In fact, it has been shown that
multivariate approaches ("mixed treatment comparisons", MTC) are most
useful when all treatments are not in all studies.

Following van Howelingen et al., I call this multivariate because there is 
an among-study variance-covariance matrix of the response values. The approach
below is used when one can assume that there is no WITHIN-study correlation
of the observations. If there is a within-study correlation (in addition to
an among-study correlation, one needs to use the more tedious approach and
create a very big PARMS data file with all the within-study (sampling) variances 
and covariances (in an exact order). 

Note: As with univariate analyses (see previous examples), one specifies
the fixed sampling variances by using a WEIGHT that is the inverse of the sampling
variance, and _also_ forces the residual variance in the model to be 1. 
But things get tricky here, because to do the latter, one must list out 
starting values for _all_ the among-study var.-cov. terms, and this number
depends on the number of groups (treatments, etc.). With three treatments,
there are six among-study var.-cov. terms (the variances for each treatment,
plus the covariances of a and b, b and c, and a and c. Thus, one must list
seven terms in the PARMS statement: the six among-study var.-cov. parameters
and the single within-study variance (forced to be 0, with the eqcons= option
[where, here, the 7th term is forced to be 1]).

It is very common for there to be very high among-study study correlations, 
resulting in a singular (or ill-conditioned) covariance matrix. Possibly
no convergence. Thus, I often replace TYPE=UN with TYPE=FA0(3) which is much
more stable, and does not blow up with high correlations. The number "3" is 
the number of groups (variables) within the studies. Just change this for
other conditions. The G and GCORR give the estimated among-study variances
and covariances, based on the parameters of the FA0(#) model for the matrix.

With the NOINT option, the F test of treat is not for treatment effect, but 
is treatment means different from 0 (not from each other). The contrast 
statement gives an overall treatment test, if this is desired.

The following is equivalent (with different syntax, with the assumption of no 
within-study covariances) to that done in van Houwelingen et al. (2002). I use a 
CLASS statement, and they used dummy variables.
*/

title2 'multiv. anal. of a, b, and c, with unstructed between cov. matrix';
proc mixed data=threestack method=reml;
class study treat;
weight samplewgt;								*<--inverse of within-study (sampling) variance;
model y = treat / chisq s noint covb corrb ddfm=bw;
random treat / sub=study type=un g gcorr;		*<--variances and covariances for all treats.;
repeated;
parms (1) (1) (1) (1) (1) (1) (1)/ eqcons=7;	*<--make sure _last_ one is fixed;
estimate 'es _ac' treat 1 -1 / cl;
estimate 'es _bc' treat 0 1 -1 / cl;
estimate 'es _ac' treat 1 0 -1 / cl;

contrast 'treat effect' treat 1 0 -1, treat -1 2 -1;
run;

/*
The statement
	random treat / sub=study type=un g gcorr;
is a generalization of the univariate 
	random int / sub=study;
This gives a separate among-study variance estimate for each treatment (group),
and a separate among-study covariance (or covariance) for each pair of treatments.
The UN type means UNstructured. G and GCORR tells SAS to print out the 
variance-covariance matrices, and corresponding correlations.

The among-study variance/covariance estimates are given in Covariance Parameter
Estimates, followed by "1" for the (single) residual variance. But because 
WEIGHTs were also used (depended on study and treatment), the latter really means
that the sampling variances were different (and fixed, because of eqcons=7) for 
each study and treatment.

Note: when fitting the univariate model to es=a-b, one gets an among-study variance
of 76.45. With the multivariate approach, one gets the following among-study terms:
	v(a)=132.29, v(b)=107.43, cov(a,b)=82.345. 
So, the among-study variance for es=a-b based on multivariate model would be: 
	132.29 + 107.43 - 2*82.345 = 75.03
which is close to the univariate-based 76.45. 

Also note that in the multivariate approach, the contrast for a-b gives:
	mean = 5.52 (se = 2.29)
and in the univariate analysis, one gets for es=a-b:
	intercept(=mean) = 5.44 (se = 2.32).
There will not (_should_ not) be complete agreement for the reasons given above 
(and for some other reasons not discussed here). But, there is generally 
good agreement.

The df calculation method for the multivariate approach is different from
that utilized within the univariate models. There is no exact correspondence.
Of course, if one uses ml, and forces large ddf (with ddf=1000), then there is no difference.
*/


*---Moderator effect in univariate approach. One _es_ example.....;

title2 'moderator effect in univariate approach, es=a-b';
proc mixed data=three;
class study moder1;
weight samplewgt;
model es = moder1 / chisq s;
random int / sub=study;
repeated;
parms (1) (1) / eqcons=2;
run;


/*
Multivariate approach for MODERATOR variables.
Note: only group (treat) and interaction (treat*moder) effects. _No_ moder main effect.
Without the moder term, the treat*moder term is a combined test of moderator effect
(main and interaction). In practice, one wants tests of moderator effects
on relevant effect sizes (a-b, etc.), rather than global tests.
One gets these from the ESTIMATEs and CONTRASTs. 
Be careful: coding is speficially for moder with two levels, and listed in
the given order in the class statement. 

The example here gives an excessive number of estimates and contrasts to 
demonstrate various things (all of these are not necessarily relevant). Others can
be used. These are examples.

The following is a work-in-progress (more could be added).
*/

title2 'moderator effect in multivariate approach';
proc mixed data=threestack;
class study treat moder;
weight samplewgt;
model y = treat  treat*moder / chisq s noint  ddfm=bw;
random treat / sub=study type=un;
repeated;
parms (1) (1) (1) (1) (1) (1) (1) / eqcons=7;

*--Now, get es (a-b, etc) for each level of moder;
estimate 'a-b @ mod=1' treat 1 -1 0 treat*moder 1 0 -1 0 0 0;
estimate 'b-c @ mod=1' treat 0 1 -1 treat*moder 0 0 1 0 -1 0;
estimate 'a-c @ mod=1' treat 1 0 -1 treat*moder 1 0 0 0 -1 0;
estimate 'a-b @ mod=2' treat 1 -1 0 treat*moder 0 1 0 -1 0 0;
estimate 'b-c @ mod=2' treat 0 1 -1 treat*moder 0 0 0 1 0 -1;
estimate 'a-c @ mod=2' treat 1 0 -1 treat*moder 0 1 0 0 0 -1;

*--Now see if es (a-b, etc.) depends on moderator;
*--These are differences of above estimates;
estimate 'moder eff on a-b' treat*moder 1 -1 -1 1 0 0 ;
estimate 'moder eff on b-c' treat*moder 0 0 1 -1 -1 1 ;
estimate 'moder eff on a-c' treat*moder 1 -1 0 0 -1 1 ;

*--repeat in contrast style;
contrast 'moder eff on a-b' treat*moder 1 -1 -1 1 0 0 ;
contrast 'moder eff on b-c' treat*moder 0 0 1 -1 -1 1 ;
contrast 'moder eff on a-c' treat*moder 1 -1 0 0 -1 1 ;

*---actual test of treatment (with noint in model, the standard F test
	is not of treatment effect;
contrast 'treat. effect' treat 1 0 -1, treat -1 2 -1;

*---the following are the same as slices;
contrast 'treat. effect @ mod=1' treat 1 0 -1 treat*moder 1 0 0 0 -1 0,
								treat -1 2 -1 treat*moder -1 0 2 0 -1 0;
contrast 'treat. effect @ mod=2' treat 1 0 -1 treat*moder 0 1 0 0 0 -1,
								treat -1 2 -1 treat*moder 0 -1 0 2 0 -1;
*---brute force calculation of the pure interaction of treatment and moderator;
*---(this does not come out of standard table, because main and interaction
	with moder are combined above;
contrast 'mod. eff. on treat. - inter.' treat*moder 1 -1 0 0 -1 1,
								treat*moder -1 1 2 -2 -1 1;
*lsmeans treat*moder / pdiff cl;
run;


/*
Now, one can use the PDATA= approach for the multivariate analysis. 
Below, this is demonstrated just for the mult. approach and _no_ moderators.
*/

data mvar1;		*<--put sampling variances (one for each study and treatment) in file;
set threestack;
est=samplingv;	*<--there are 45 variances (15 studies * 3 treatments);
keep study est;
data mvar2;		*<--put initial guesses of among-study variances and covariances in file;
input est @@;	*<--there are six with three treatments;
datalines;
1 1 1 1 1 1
;
data mvariance; set mvar2 mvar1;		*<--stack files;

title2 'intial guesses of among-study variances/covariances, and sampling variances';
proc print data=mvariance;
run;
title2 'multiv. anal. of a, b, and c,  unstructed between cov. matrix';
title3 'no moderators';
proc mixed data=threestack method=reml;
class study treat;								
model y = treat / chisq s noint covb corrb ddfm=bw;
random treat / sub=study type=un g gcorr;
repeated / group = study*treat type=simple;
parms / pdata=mvariance eqcons=7 to 51;		*<--make sure _last_ 45 (15*3) are fixed;
estimate 'es _ac' treat 1 -1 / cl;
estimate 'es _bc' treat 0 1 -1 / cl;
estimate 'es _ac' treat 1 0 -1 / cl;
contrast 'treat effect' treat 1 0 -1, treat -1 2 -1;
run;
/*
	In above, instead of using a within-study WEIGHT (that is different for each study)
	and a within-study residual variance fixed at 1, 
	one gives the within-study sampling variances in the mvarinace file 
	(after first giving guesses for the among-study variances and coviances).

	With this formulation, one needs the repeated statement 
 		repeated / group = study*treat type=simple;
	to indicate that there is a different sampling variance for each study and treatment.

	Under Covariance Parameter Estimates in the OUTPUT, the among-study variance
	and covariance estimates are given first, followed by the 45 (15*3) fixed
	sampling varainces. Results are identical to the first mult. analysis, based
	on WEIGHT. 

	For other MULTIVARIATE settings, such as two or more correlated random variables
	in each study, one must use the PDATA= approach to give the within-study
	sampling variances and covariances for each study. This is _tricky_ (not shown here).
*/

run;

*ods html close;
run;
